# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.modules.backups"""

import logging
import os
import threading
import time
import uuid

from datetime import datetime

import six

from boto3.s3 import transfer as S3Transfer

from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.errors import DWhoConfigurationError
from dwho.classes.modules import DWhoModuleBase, MODULES
from dwho.classes.inotify import DWhoInotifyConfig
from dwho.classes.notifiers import DWhoPushNotifications
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe.libs import anysql, mysql_config_parser, urisup
from sonicprobe.libs.moresynchro import RWLock

from ..classes.s3 import (XtrargonS3Helper,
                          XtrargonTxProgress,
                          XTRARGON_FILE)


LOG                         = logging.getLogger('xtrargon_ng.modules.backups')

DATE_FORMAT                 = '%Y-%m-%d %H:%M:%S'
DEFAULT_OBJECT_TTL          = 604800
DEFAULT_DB_SLEEP_DURATION   = 0.5
DEFAULT_MAX_UPLOAD_ATTEMPTS = 10
DEFAULT_SAFE_SLAVE_TIMEOUT  = 300

DB_OPEN_TEMP_TABLES_WAIT    = 3
DB_LOCK_WAIT_TIMEOUT        = 31536000

STATUS_FAILURE              = 'failure'
STATUS_PROCESSING           = 'processing'
STATUS_SUCCESS              = 'success'
STATUS_STOPPED              = 'stopped'
STATUS_UNKNOWN              = 'unknown'


class BackupsLockDB(threading.Thread):
    def __init__(self, module):
        threading.Thread.__init__(self)
        self.name                = 'lockdb'
        self._module             = module
        self._config             = module.config['backups'].get('db') or {}
        self._lock_db            = module.lock_db
        self._db_locked          = module.db_locked
        self._uri                = module.dburi

        self._killed             = False

        self._dbconn             = None
        self._dbcursor           = None
        self._sleep_duration     = float(DEFAULT_DB_SLEEP_DURATION)
        self._sql_thread_run     = False
        self._no_lock            = bool(self._config.get('no_lock', False))
        self._safe_slave         = bool(self._config.get('safe_slave', True))
        self._safe_slave_timeout = int(self._config.get('safe_slave_timeout', DEFAULT_SAFE_SLAVE_TIMEOUT))
        if self._safe_slave_timeout < 1:
            self._safe_slave_timeout = DEFAULT_SAFE_SLAVE_TIMEOUT

    def _stop(self):
        if self._dbcursor:
            LOG.info("DB UNLOCK TABLES")
            self._dbcursor.query("UNLOCK TABLES")
            self._dbcursor = None

        if self._dbconn:
            self._dbconn.close()
            self._dbconn = None

    def _check_connection(self):
        if not self._dbconn:
            raise LookupError("dbconn uninitialized")

        if not self._dbconn.is_connected():
            raise LookupError("dbconn unconnected")

        if not self._dbcursor:
            raise LookupError("dbcursor uninitialized")

    def _get_sql_variable(self, varname):
        self._dbcursor.query("SHOW VARIABLES LIKE '%s'" % varname)
        row = self._dbcursor.fetchone(raw = True)
        if row:
            return row['Value']

        return None

    def _get_slave_status(self):
        self._dbcursor.query("SHOW SLAVE STATUS")
        row = self._dbcursor.fetchone(raw = True)
        if row:
            return row

        return None

    def _check_is_slave(self):
        if self._killed:
            raise GeneratorExit()

        self._check_connection()

        server_id = self._get_sql_variable('server_id')
        if not server_id or int(server_id) == 0:
            return False

        slave_status = self._get_slave_status()
        if not slave_status \
           or not slave_status['Read_Master_Log_Pos'] \
           or not slave_status['Slave_SQL_Running']:
            return False

        self._sql_thread_run = slave_status['Slave_SQL_Running'] == 'Yes'

        return slave_status['Slave_SQL_Running']

    def _get_slave_coordinates(self):
        slave_status = self._get_slave_status()
        if not slave_status:
            return None

        return (slave_status['Relay_Master_Log_File'],
                slave_status['Exec_Master_Log_Pos'])

    def _get_slave_open_temp_tables(self):
        open_temp_tables = self._get_sql_variable('Slave_open_temp_tables')
        if not open_temp_tables:
            return 0

        return int(open_temp_tables)

    def _check_slave_open_temp_tables(self, start_time):
        prev_slave_coordinates = None
        curr_slave_coordinates = self._get_slave_coordinates()
        open_temp_tables       = None

        while True:
            if self._killed:
                raise GeneratorExit()

            open_temp_tables = self._get_slave_open_temp_tables()
            if not open_temp_tables:
                LOG.info("no open temp tables")
                break

            if start_time + self._safe_slave_timeout > time.time():
                LOG.info("time exceeded for open temp tables")
                return False

            prev_slave_coordinates = curr_slave_coordinates

            self._dbcursor.query("START SLAVE SQL_THREAD")
            time.sleep(DB_OPEN_TEMP_TABLES_WAIT)

            curr_slave_coordinates = self._get_slave_coordinates()
            if prev_slave_coordinates \
               and curr_slave_coordinates \
               and prev_slave_coordinates == curr_slave_coordinates:
                LOG.info("Slave pos hasn't moved during wait period, "
                         "not stopping the SQL thread.")
            else:
                self._dbcursor.query("STOP SLAVE SQL_THREAD")

        return open_temp_tables

    def _wait_for_safe_slave(self):
        start_time = time.time()

        if not self._check_is_slave():
            LOG.info("db is not a slave")
            return None

        try:
            if self._sql_thread_run:
                self._dbcursor.query("STOP SLAVE SQL_THREAD")

            while True:
                if self._killed:
                    raise GeneratorExit()

                self._check_connection()

                open_temp_tables = self._check_slave_open_temp_tables(start_time)
                if open_temp_tables is False:
                    return False

                if open_temp_tables == 0:
                    self._dbcursor.query("STOP SLAVE SQL_THREAD")
                    open_temp_tables = self._get_slave_open_temp_tables()
                    if open_temp_tables != 0:
                        continue
                return True
        finally:
            if self._sql_thread_run:
                self._dbcursor.query("START SLAVE SQL_THREAD")
            else:
                self._dbcursor.query("STOP SLAVE SQL_THREAD")

    def _acquire_lock(self):
        if self._killed:
            raise GeneratorExit()

        if self._db_locked.is_set():
            return

        self._check_connection()

        if self._safe_slave:
            self._wait_for_safe_slave()

        if not self._no_lock:
            if self._get_sql_variable('lock_wait_timeout') is not None:
                self._dbcursor.query("SET SESSION lock_wait_timeout=%d" % DB_LOCK_WAIT_TIMEOUT)

            if self._get_sql_variable('wresp_on') in ('1', 'ON'):
                self._dbcursor.query("SET SESSION wsrep_causal_reads=0")

            self._dbcursor.query("FLUSH NO_WRITE_TO_BINLOG TABLES")
            self._dbcursor.query("FLUSH TABLES WITH READ LOCK")

        self._db_locked.set()

    def _alive(self):
        while not self._killed and self._db_locked.is_set():
            if not self._lock_db.is_set():
                return

            self._check_connection()

            LOG.info("DB SLEEP(%s)", self._sleep_duration)
            self._dbcursor.query("SELECT SLEEP(%s)" % self._sleep_duration)

    def run(self):
        self._lock_db.wait()
        LOG.info("starting db lock")

        try:
            if self._killed:
                raise GeneratorExit()

            self._dbconn = anysql.connect_by_uri(self._uri)
            self._dbcursor = self._dbconn.cursor()
            self._acquire_lock()
            self._alive()
        except GeneratorExit:
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping db lock")
        except Exception as e:
            self._module.add_errors(repr(e))
            LOG.exception(e)
        finally:
            self._lock_db.clear()
            self._db_locked.clear()
            self._stop()

    def terminate(self):
        LOG.info("db lock terminated")
        self._killed = True
        self._lock_db.set()
        self._db_locked.set()
        self._stop()


class BackupsScanThread(threading.Thread):
    def __init__(self, module):
        threading.Thread.__init__(self)
        self.name           = 'scan'
        self._module        = module
        self._config        = module.config
        self._s3helper      = module.s3helper

        self._step          = 1
        self._max_attempts  = 1
        self._killed        = False

        self._bucket        = self._s3helper.get_bucket()
        self._storage_class = self._s3helper.get_storage_class()
        self._client        = self._s3helper.boto3_client()
        self._object_ttl    = DEFAULT_OBJECT_TTL
        self._file_sync     = module.file_sync

        if 'object_ttl' in self._config['backups']:
            self._object_ttl = int(self._config['backups']['object_ttl'])

    @staticmethod
    def _mk_sum(filepath):
        if not os.path.exists(filepath):
            return ''

        fstat = os.stat(filepath)

        return '|'.join((str(fstat.st_size),
                         str(fstat.st_mtime)))

    def _mk_filekey(self, filepath, xid = None):
        if xid:
            return "%s:%s:file:%s" % (self._module.server_id, xid, filepath)

        return "%s:%s:file:%s" % (self._module.server_id, xid, filepath)

    def _s3up(self, src, dst = None):
        if not dst:
            dst = src

        if not os.path.exists(src):
            return None

        last_exception = None

        max_attempts = 1
        if self._step > 1:
            max_attempts = self._max_attempts

        for i in range(max_attempts):
            try:
                with S3Transfer.S3Transfer(self._s3helper.boto3_client(),
                                           self._s3helper.transfer_config()) as transfer:
                    transfer.upload_file(
                        filename   = src,
                        bucket     = self._bucket,
                        key        = dst.lstrip('/'),
                        callback   = XtrargonTxProgress(src, False, self._killed),
                        extra_args = {'StorageClass': self._storage_class})

                LOG.info("s3 upload (src: %r, dst: %r, bucket: %r, attempt: %s / %s)",
                         src,
                         dst,
                         self._bucket,
                         i,
                         max_attempts)
                return True
            except Exception as e:
                LOG.debug("Retrying exception caught (%s), "
                          "retrying request, (attempt %s / %s)", e, i,
                          max_attempts, exc_info=True)
                last_exception = e
                continue

        if self._step > 1:
            raise Exception(last_exception)

    def _s3api(self, cmd, key = None, **kwargs):
        kwargs['Bucket'] = self._bucket
        if key:
            kwargs['Key'] = key.lstrip('/')
        kwargs['StorageClass'] = self._storage_class

        LOG.info("s3 api: %r: %r", cmd, kwargs)
        getattr(self._client, cmd)(**kwargs)

    def _touch(self, xpath, force = False):
        if not force and self._file_sync.exists(xpath):
            return

        self._s3api('put_object', xpath, **{'Body': ''})

        self._file_sync.set_key(xpath, '', self._object_ttl)

    def _upload(self, srcpath, destpath, force = False):
        xid  = self._module.get_state_value('started_at')
        fkey = self._mk_filekey(srcpath, xid)
        nsum = self._mk_sum(srcpath)

        if force:
            self._s3up(srcpath, destpath)
            self._file_sync.set_key(fkey, nsum, self._object_ttl)
            return

        osum = self._file_sync.get_key(fkey)
        if not osum or osum != nsum:
            self._s3up(srcpath, destpath)

        self._file_sync.set_key(fkey, nsum, self._object_ttl)

    @staticmethod
    def _destpath(srcpath, params, filepath, xvars):
        destpath = filepath
        fkwargs  = xvars.copy()
        fkwargs['filepath'] = filepath
        fkwargs['dirname']  = os.path.dirname(filepath)
        fkwargs['filename'] = os.path.basename(filepath)

        if not params or not isinstance(params, dict):
            return destpath

        if params.get('format'):
            destpath = params['format'].format(**fkwargs)

        if params.get('subst'):
            if isinstance(params['subst'], six.string_types):
                xsubst = params['subst']
            else:
                xsubst = srcpath

            if filepath.startswith(xsubst):
                substpath = filepath[len(xsubst):].strip(os.path.sep)
                if params.get('format'):
                    destpath = os.path.join(os.path.sep, destpath.strip(os.path.sep), substpath)
                else:
                    destpath = os.path.join(os.path.sep, substpath)

        if params.get('prefix'):
            destpath = os.path.join(params['prefix'].strip(os.path.sep),
                                    destpath.strip(os.path.sep))

        return os.path.join(os.path.sep, destpath.strip(os.path.sep))

    def _walk(self, srcpath, params, xvars):
        for root, dirs, files in os.walk(srcpath, topdown = True, followlinks = params['follow']):
            if self._killed:
                raise GeneratorExit()

            if params['exclude_patterns'] and params['exclude_patterns'](root):
                dirs[:] = []
                LOG.debug("exclude path from scan. (path: %r)", root)
                continue

            self._max_attempts = int(params.get('max_upload_attempts', DEFAULT_MAX_UPLOAD_ATTEMPTS) or 1)
            if self._max_attempts < 1:
                self._max_attempts = 1

            if not dirs and not files:
                filepath = os.path.join(root, XTRARGON_FILE)
                destpath = self._destpath(srcpath, params.get('dest_params'), filepath, xvars)
                self._touch(destpath)
                continue

            for filename in files:
                if self._killed:
                    raise GeneratorExit()

                filepath = os.path.join(root, filename)
                destpath = self._destpath(srcpath, params.get('dest_params'), filepath, xvars)

                if params['exclude_patterns'] and params['exclude_patterns'](filepath):
                    LOG.debug("exclude file from scan. (filepath: %r)", filepath)
                    continue

                self._upload(filepath, destpath)

    def _run(self, fkwargs):
        for srcpath, params in six.iteritems(self._config['backups']['paths']):
            if self._killed:
                raise GeneratorExit()

            self._walk(srcpath, params, fkwargs)

    def _post_run(self):
        if self._config['backups'].get('notifications'):
            self._module.notifiers(self._module.get_state(end = True))

    def run(self):
        self._module.is_running.wait()
        LOG.info("starting backup")
        self._module.set_state_value('started_at', time.strftime(DATE_FORMAT))
        self._module.set_state_value('uuid', "%s" % uuid.uuid4())
        self._module.set_state_value('status', STATUS_PROCESSING)
        self._module.is_scanning.set()

        fkwargs = {'env': os.environ,
                   'gmtime': datetime.utcnow(),
                   'time': datetime.now(),
                   'server_id': self._module.server_id}

        try:
            if self._killed:
                raise GeneratorExit()

            LOG.info("first step of backup")
            self._step = 1
            self._run(fkwargs)
            self._module.lock_db.set()
            self._module.db_locked.wait()

            if self._killed:
                raise GeneratorExit()

            LOG.info("second step of backup")
            self._step = 2
            self._run(fkwargs)
        except GeneratorExit:
            self._module.set_state_value('status', STATUS_STOPPED)
            LOG.info("stopping backup")
        except Exception as e:
            self._module.add_errors(repr(e))
            LOG.exception(e)
        finally:
            self._module.lock_db.clear()
            self._module.is_scanning.clear()
            self._module.is_running.clear()

        self._post_run()

    def terminate(self):
        LOG.info("backup terminated")
        self._killed = True
        self._module.is_running.set()
        self._module.is_scanning.set()


class BackupsModule(DWhoModuleBase):
    MODULE_NAME = 'backups'

    LOCK = RWLock()

    @staticmethod
    def _load_path_params(params):
        r = {'follow': True,
             'exclude_patterns': None,
             'dest_params': None}

        if not params:
            return r

        r['follow'] = bool(params.get('follow', True))

        if 'exclude_files' in params:
            r['exclude_patterns'] = DWhoInotifyConfig.load_exclude_patterns(params['exclude_files'])

        if isinstance(params['dest_params'], dict):
            r['dest_params'] = params['dest_params'].copy()

        return r

    def _load_s3helper(self):
        if not self.config.get('backups'):
            raise DWhoConfigurationError("missing backups configuration")

        for x in ('paths', 's3'):
            if x not in self.config['backups'] \
             or not isinstance(self.config['backups'][x], dict):
                raise DWhoConfigurationError("missing backups %s" % x)

        for srcpath, params in six.iteritems(self.config['backups']['paths'].copy()):
            self.config['backups']['paths'][srcpath] = self._load_path_params(params)

        r = XtrargonS3Helper(self.config['backups']['s3'])
        r.load_credentials()

        if not r.get_bucket():
            raise DWhoConfigurationError("missing default bucket setting")

        return r

    def _build_mysql_uri(self):
        host = mysql_config_parser.MYSQL_DEFAULT_HOST
        port = mysql_config_parser.MYSQL_DEFAULT_PORT

        if self.dbconf.has_option('client', 'host'):
            host = self.dbconf.get('client', 'host')

        if self.dbconf.has_option('client', 'port'):
            port = self.dbconf.get('client', 'port')

        uri = ('mysql',
               (self.dbconf.get('client', 'user'),
                self.dbconf.get('client', 'password'),
                host,
                port),
               None,
               (('cursorclass', 'Dict'),),
               None)

        return urisup.uri_help_unsplit(uri)

    @staticmethod
    def _check_mysql_server(uri):
        dbconn = None
        try:
            dbconn = anysql.connect_by_uri(uri)
            if not dbconn.is_connected():
                raise DWhoConfigurationError("unable to connect to mysql server")
        finally:
            if dbconn:
                dbconn.close()

    def _check_redis_server(self, prefix):
        redisconn = DWhoAdapterRedis(self.config, prefix = prefix)
        if not redisconn.servers:
            raise DWhoConfigurationError("missing redis server for %s" % prefix)

        redisconn.ping()

        return redisconn

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']

        self.is_running   = threading.Event()
        self.is_scanning  = threading.Event()
        self.lock_db      = threading.Event()
        self.db_locked    = threading.Event()

        self._bkp_scan    = None
        self._db_lock     = None
        self._errors      = []
        self._state       = {}

        self.dbconf       = mysql_config_parser.MySQLConfigVersion(
            custom_file = os.path.join('client.d', "%s.cnf" % self.server_id),
            config_dir  = self.config.get('_config_directory')).get_client()
        self.dburi        = self._build_mysql_uri()

        self._check_mysql_server(self.dburi)
        self._bkp_history = self._check_redis_server('backups_history')
        self.file_sync    = self._check_redis_server('file_sync')

        self.s3helper     = self._load_s3helper()

        self.notifiers    = None
        if self.config['general'].get('notifiers_path'):
            self.notifiers = DWhoPushNotifications(self.server_id,
                                                   self.config['general']['notifiers_path'])

    def _stop(self):
        if self._bkp_scan:
            self._bkp_scan.terminate()
            self._bkp_scan = None

        if self._db_lock:
            self._db_lock.terminate()
            self._db_lock = None

    def at_stop(self):
        self._stop()

    def _start(self, start = True, clear_events = True):
        self._stop()
        self.reset_state()

        self._db_lock  = BackupsLockDB(self)
        self._db_lock.daemon = True

        self._bkp_scan = BackupsScanThread(self)
        self._bkp_scan.daemon = True

        if clear_events:
            self.is_running.clear()
            self.is_scanning.clear()
            self.lock_db.clear()
            self.db_locked.clear()

        if start:
            self._bkp_scan.start()
            self._db_lock.start()
            self.is_running.set()

    def add_errors(self, errors):
        if isinstance(errors, six.string_types):
            errors = [errors]

        if isinstance(errors, list):
            self._errors.extend(errors)

    def has_errors(self):
        return len(self._errors) > 0

    def get_state(self, status = None, end = False):
        self._state['errors'] = self._errors

        if status:
            self._state['status'] = status

        if not end:
            return self._state

        started_at = self._state.get('started_at')
        if started_at:
            started_at = datetime.strptime(started_at, DATE_FORMAT)
            duration   = (datetime.now() - started_at).total_seconds()
            self._state['duration'] = duration

        self._state['ended_at'] = time.strftime(DATE_FORMAT)

        if self._state['status'] == STATUS_STOPPED:
            self._state['errors'].append('stopped before completing')
        elif not self.has_errors():
            self._state['status'] = STATUS_SUCCESS
        else:
            self._state['status'] = STATUS_FAILURE

        return self._state

    def reset_state(self):
        self._state = {'status':     STATUS_UNKNOWN,
                       'started_at': None,
                       'ended_at':   None,
                       'uuid':       None,
                       'duration':   -1,
                       'errors':     []}

    def get_state_value(self, key):
        return self._state.get(key)

    def set_state_value(self, key, value):
        self._state[key] = value

    def start(self, request): # pylint: disable-msg=unused-argument
        if self.is_running.is_set():
            raise HttpReqErrJson(409, "backup is already processing")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            self._start()
            self.is_scanning.wait(self.lock_timeout)
            return self.get_state()
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()

    def stop(self, request): # pylint: disable-msg=unused-argument
        if not self.is_running.is_set():
            raise HttpReqErrJson(409, "backup isn't running")

        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            self._stop()
            xtime = time.time()

            while self.is_running.is_set():
                if time.time() - xtime > self.lock_timeout:
                    self.add_errors("unable to stop backup")
                    break
                time.sleep(0.1)
            return self.get_state(end = True)
        except HttpReqErrJson:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(BackupsModule())
    _start()
